package com.ldodds.sparql;

import java.io.*;

import com.hp.hpl.jena.util.*;
import com.hp.hpl.jena.query.*;
import com.hp.hpl.jena.query.core.ARQInternalErrorException;
import com.hp.hpl.jena.query.vocabulary.ResultSetVocab;
import com.hp.hpl.jena.rdf.model.*;

import org.apache.commons.logging.*;

/**
 * 
 * @author Andy Seaborne
 * @author Leigh Dodds
 */
public class RunQuery
{
    /** No output */
    public final static String FMT_NONE = "none";

    /** Output the triples for the query */
    public final static String FMT_TUPLES = "tuples";

    /** Output a text table */
    public final static String FMT_TEXT = "text";

    /** Output an RDF graph */    
    public final static String FMT_RS_RDF = "rs/graph";

    /** Output results showing structure */
    public final static String FMT_RS_TEXT = "rs/text";

    /** Output as XML */
    public final static String FMT_RS_XML = "rs/xml";

    private Log _log = LogFactory.getLog(RunQuery.class);
    private OutputStream _out;
    private QueryContext _context;

    public RunQuery(QueryContext context)
    {
        this(System.out, context);
    }

    public RunQuery(OutputStream out, QueryContext context)
    {
        _out = out;
        _context = context;
    }

    /** 
     * Execute a query given some information.
     * Set what you know in the QueryCntext object - this code will
     * sort out missing stuff if it can.
     *    
     * @param _context
     */
    public void query()
    {
        addDataSourceToContext();

        createQuery();
        
        if (_context.query == null) return;

        bindData();

        // Check there is a dataset
        if (_context.query.getDataSet() == null
                && _context.query.getSourceURIs() == null
                && _context.query.getNamedSourceURIs() == null)

        {
            //TODO report errors in GUI
            System.err
                    .println("Dataset not specified in query nor provided on command line.");
            return;
        }
        execute();
    }

    private void addDataSourceToContext()
    {
        try
        {
            if (_context.fileManager == null) _context.fileManager = FileManager.get();
            if (_context.dataURL != null)
                _context.model = _context.fileManager.loadModel(_context.dataURL,
                        _context.baseURI, _context.dataSyntax);
        }
        catch (Exception ex)
        {
            //TODO report errors in GUI            
            System.err.println("Error loading URL " + _context.dataURL);
            ex.printStackTrace();
        }
    }

    /**
     * Reparse query if necessary, and associate baseURI.
     */
    void createQuery()
    {
        // Create the query
        if (_context.query == null)
        {
            _context.query = parseQuery(_context.queryString, _context.syntax);
            if (_context.query == null)
            {
                // Errors already output
                return;
            }
        }
        if (_context.baseURI != null) 
        {
            _context.query.setBaseURI(_context.baseURI);
        }
    }

    private void bindData()
    {
        // Bind up the data
        if (_context.model != null) _context.query.setDataSet(_context.model);
    }

    private void execute()
    {
        try
        {
            // Execute
            QueryExecution qe = QueryFactory.createQueryExecution(_context.query);
            if (_context.fileManager != null)
            {
                qe.setFileManager(_context.fileManager);
            }
            qe.init();

            doQuery(_context.query, qe, _context.outputFormat);
            qe.close();
        }
        catch (ARQInternalErrorException intEx)
        {
            //TODO report errors in GUI
            System.err.println(intEx.getMessage());
            if (intEx.getCause() != null)
            {
                System.err.println("Cause:");
                intEx.getCause().printStackTrace(System.err);
                System.err.println();
            }
            intEx.printStackTrace(System.err);
        }
        catch (QueryException qEx)
        {
            //TODO report errors in GUI
            qEx.printStackTrace();
        }
        catch (Exception ex)
        {
            //TODO report errors in GUI
            ex.printStackTrace();
        }
    }

    private Query parseQuery(String queryString, String syntax)
    {
        try
        {
            Query query = Query.create(queryString, syntax);
            return query;
        }
        catch (ARQInternalErrorException intEx)
        {
            System.err.println(intEx.getMessage());
            if (intEx.getCause() != null)
            {
                //TODO report errors in GUI
                System.err.println("Cause:");
                intEx.getCause().printStackTrace(System.err);
                System.err.println();
            }
            intEx.printStackTrace(System.err);
            return null;
        }
        catch (QueryException qEx)
        {
            //TODO report errors in GUI            
            System.err.println(qEx.getMessage());
            return null;
        }
        catch (Exception ex)
        {
            //TODO report errors in GUI            
            System.out.flush();
            ex.printStackTrace(System.err);
            return null;
        }
    }

    private void doQuery(Query query, QueryExecution qe,
            String outputFormat)
    {
        if (query.isSelectType()) doSelectQuery(query, qe, outputFormat);
        if (query.isDescribeType()) doDescribeQuery(query, qe, outputFormat);
        if (query.isConstructType()) doConstructQuery(query, qe, outputFormat);
        if (query.isAskType()) doAskQuery(query, qe, outputFormat);
    }

    private void doSelectQuery(Query query, QueryExecution qe,
            String outputFormat)
    {
        ResultSet results = qe.execSelect();

        // Force query to execute - until now it has merely been set up.
        results = ResultSetFactory.rewindableResults(results);

        boolean done = false;

        // The non-display forms / uses a ResultSetFormatter 

        if (outputFormat.equals(FMT_NONE))
        {
            ResultSetFormatter fmt = new ResultSetFormatter(results, query
                    .getPrefixMap());
            fmt.consume();
            fmt.close();
            done = true;
        }

        if (outputFormat.equals(FMT_RS_RDF))
        {
            ResultSetFormatter rsFmt = new ResultSetFormatter(results, query
                    .getPrefixMap());
            Model m = rsFmt.toModel();
            RDFWriter rdfw = m.getWriter("TURTLE");
            m.setNsPrefix("rs", ResultSetVocab.getURI());
            rdfw.write(m, _out, null);
            rsFmt.close();
            done = true;
        }

        PrintWriter out = FileUtils.asPrintWriterUTF8(_out);

        if (outputFormat.equals(FMT_RS_XML))
        {
            ResultSetFormatter rsFmt = new ResultSetFormatter(results, query
                    .getPrefixMap());
            rsFmt.outputAsXML(_out);
            rsFmt.close();
            done = true;
        }

        if (outputFormat.equals(FMT_RS_TEXT))
        {
            ResultSetFormatter rsFmt = new ResultSetFormatter(results, query
                    .getPrefixMap());
            rsFmt.outputAsText(_out);
            done = true;
        }

        ResultSetTextFormatter fmt = null;

        if (outputFormat.equals(FMT_TEXT))
        {
            fmt = new ResultSetTextFormatter(results, query.getPrefixMap());
            fmt.printAll(out);
            done = true;
        }

        if (outputFormat.equals(FMT_TUPLES))
        {
            fmt = new ResultSetTextFormatter(results);
            fmt.dump(out, true);
            done = true;
        }

        if (!done) _log.warn("Unknown format request: " + outputFormat);

        if (out != null) out.flush();

        if (fmt != null) fmt.close();

        results.close();

        if (query.getQueryBlock().getDataSource() != null)
            query.getQueryBlock().getDataSource().close();
    }

    private void doDescribeQuery(Query query, QueryExecution qe,
            String outputFormat)
    {
        Model r = qe.execDescribe();
        PrintWriter out = FileUtils.asPrintWriterUTF8(_out);
        
        if (!outputFormat.equals(FMT_NONE))
        {
            out.println("======== DESCRIBE results ");
            r.write(out, "N3", null);
        }
        out.println("======== ");
    }

    private void doConstructQuery(Query query, QueryExecution qe,
            String outputFormat)
    {
        Model r = qe.execConstruct();
        PrintWriter out = FileUtils.asPrintWriterUTF8(_out);
        
        if (!outputFormat.equals(FMT_NONE))
        {
            out.println("======== CONSTRUCT results ");
            r.write(System.out, "N3", null); // Base is meaningless
            out.println("======== ");
        }
    }

    private void doAskQuery(Query query, QueryExecution qe, String outputFormat)
    {
        boolean b = qe.execAsk();
        PrintWriter out = FileUtils.asPrintWriterUTF8(_out);
        if (!outputFormat.equals(FMT_NONE))
        {
            out.println("Ask => " + (b ? "Yes" : "No"));
        }
    }

}